'''
Created on 2021-05-03

Proceedings Title signature elements
@author: wf
'''
from ptp.parsing import Category
from num2words import num2words
import re
import yaml
import os
from pyparsing import ParseException

class RegexpCategory(Category):
    '''
    category defined by regular expression
    '''
    def __init__(self,name,itemFunc,pattern):
        self.pattern=pattern
        super().__init__(name,itemFunc=itemFunc)
        
    def checkMatch(self,token):
        return re.search(self.pattern, token) is not None
    
class ParsingCategory(Category):
    '''
    category defined by a PyParsing grammar
    '''
    def __init__(self,name,grammar):
        self.grammar=grammar
        itemFunc=self.getResult
        super().__init__(name,itemFunc=itemFunc)
    
    def getResult(self):
        pass
        return None
    
    def checkMatch(self,token):
        self.parseResult=None
        try:
            self.parseResult=self.grammar.parseString(token)
            pass
        except ParseException as pe:
            pass
        return self.parseResult is not None
        
  
class EnumCategory(Category):
    tokens=None
    
    @classmethod
    def getYamlPath(cls):
        path=os.path.dirname(__file__)+"/../resources"
        path=f"{path}/dictionary.yaml"
        return path

    @classmethod    
    def read(cls, yamlPath=None):
        ''' read the dictionary from the given yaml path
            https://github.com/WolfgangFahl/ProceedingsTitleParser/blob/7e52b4e3eae09269464669fe387425b9f6392952/ptp/titleparser.py#L456
        '''
        if yamlPath is None:
            yamlPath=EnumCategory.getYamlPath()
        with open(yamlPath, 'r') as stream:
            cls.tokens = yaml.safe_load(stream)
        pass
    
    def __init__(self,name):
        '''
        construct me for the given name
        '''
        super().__init__(name,itemFunc=lambda word:self.lookup(word))
        self.lookupByKey={}    
        if EnumCategory.tokens is None:
            EnumCategory.read()
        for tokenKey in EnumCategory.tokens:
            token=EnumCategory.tokens[tokenKey]
            tokenType=token['type']
            if tokenType==name:
                value=token['value'] if 'value' in token else tokenKey 
                self.lookupByKey[tokenKey]=value
    
    def checkMatch(self,word):
        return self.lookup(word) is not None
    
    def lookup(self,word):
        if word in self.lookupByKey:
                return self.lookupByKey[word]
        return None
    
    def addLookup(self,key,value):
        '''
        add the given key value pair to my lookup
        '''
        self.lookupByKey[key]=value
    
class OrdinalCategory(EnumCategory):
    '''
    I am the category for ordinals
    '''

    def __init__(self,maxOrdinal=250):
        '''
        Constructor
        '''
        self.maxOrdinal=maxOrdinal
        super().__init__("Ordinal")
        self.prepareLookup()
        
    def prepareLookup(self):
        # https://stackoverflow.com/a/20007730/1497139
        ordinal = lambda n: "%d%s" % (n,"tsnrhtdd"[(n//10%10!=1)*(n%10<4)*n%10::4])
        for i in range(1,self.maxOrdinal):
            # standard decimal ordinal 1., 2., 3. 
            self.addLookup(f"{i}.",i)
            # text ordinal 1st, 2nd, 3rd
            ordinalStr=f"{ordinal(i)}"
            self.addLookup(ordinalStr,i)
            # roman
            self.addLookup(f"{self.toRoman(i)}.",i)
            # using num2words library
            ordinal4i=num2words(i, to='ordinal')
            self.addLookup(ordinal4i,i)
            title=ordinal4i.title()
            self.addLookup(title,i)
            pass       
    
    def toRoman(self,number):
        ''' https://stackoverflow.com/a/11749642/1497139 '''
        if (number < 0) or (number > 3999):
            raise Exception("number needs to be between 1 and 3999")
        if (number < 1): return ""
        if (number >= 1000): return "M" + self.self.toRoman(number - 1000)
        if (number >= 900): return "CM" + self.toRoman(number - 900)
        if (number >= 500): return "D" + self.toRoman(number - 500)
        if (number >= 400): return "CD" + self.toRoman(number - 400)
        if (number >= 100): return "C" + self.toRoman(number - 100)
        if (number >= 90): return "XC" + self.toRoman(number - 90)
        if (number >= 50): return "L" + self.toRoman(number - 50)
        if (number >= 40): return "XL" + self.toRoman(number - 40)
        if (number >= 10): return "X" + self.toRoman(number - 10)
        if (number >= 9): return "IX" + self.toRoman(number - 9)
        if (number >= 5): return "V" + self.toRoman(number - 5)
        if (number >= 4): return "IV" + self.toRoman(number - 4)
        if (number >= 1): return "I" + self.toRoman(number - 1)


'''
Created on 2022-03-06

@author: wf
'''
from datetime import datetime
from corpus.utils.progress import Progress
from geograpy.locator import City
from lodstorage.tabulateCounter import TabulateCounter
from collections import Counter
from geograpy.locator import Locator
from corpus.location import LocationLookup
from corpus.lookup import CorpusLookup

class LocationFixer(object):
    '''
    fixer for locations
    '''
    
    def __init__(self):
        '''
        constructor
        '''
        # makes sure the geograpy3 data is available
        Locator.resetInstance()
        locator=Locator.getInstance()
        locator.downloadDB()
        self.locationLookup=LocationLookup()      
    
    def getCounter(self,events:list,propertyName:str):
        '''
        get a counter for the given propertyName
        '''
        counter=Counter()
        for event in events:
            if hasattr(event,propertyName):
                value=getattr(event,propertyName)
                if value is not None:
                    counter[value]+=1
        tabCounter=TabulateCounter(counter)
        return counter,tabCounter
    
    def timeStamp(self):
        '''
        get a timestamp with second precision
        '''
        return datetime.utcnow().strftime('%Y-%m-%d_%H:%M:%S')
    
    def fixLocations4LookupId(self,lookupIds,logFileRoot="/tmp"):
        '''
        fix locations 4 the given lookup ids
        '''
        self.lookup=CorpusLookup(lookupIds=lookupIds)
        self.lookup.load(forceUpdate=False)
        for lookupId in lookupIds:
            dataSource=self.lookup.getDataSource(lookupId)
            eventManager=dataSource.eventManager
            show=True
            locationAttribute=dataSource.sourceConfig.locationAttribute
            limit=len(eventManager.getList())
            addLocationInfo=True
            showProgress=True
            logFile=None
            if logFileRoot is not None:
                timestamp=self.timeStamp()
                logFilePath=f"{logFileRoot}/{lookupId}-{timestamp}.log"
                logFile = open(logFilePath,'w')
            if lookupId=="dblp":
                eventManager.addLocations()
            self.fixLocations(eventManager, locationAttribute, addLocationInfo, limit, showProgress=showProgress,logFile=logFile)

    def fixLocations(self,eventManager,locationAttribute,addLocationInfo=False,limit=100,showProgress=True,logFile=None):
        '''
        fix locations
        
        Args:
            eventManager: the eventmanager to use
            locationAttribute(str): the name of the location attribute
            addLocationInfo(bool): if True add the location information
        '''
        events=eventManager.events
        pCount,_pCountTab=self.getCounter(events,locationAttribute)
        eventsByLocation=eventManager.getLookup(locationAttribute,withDuplicates=True)
        count=len(pCount.items())
        total=sum(pCount.values())
        rsum=0
        problems=[]
        progress=Progress(progressSteps=1,expectedTotal=total,msg=f"Fixing {limit}/{total} locations for {eventManager.name} using location attribute {locationAttribute}",showMemory=True)
        for i,locationTuple in enumerate(pCount.most_common(limit)):
            locationText,locationCount=locationTuple
            rsum+=locationCount
            percent=rsum/total*100
            city=None
            try:
                city=self.locationLookup.lookup(locationText,logFile=logFile)
            except Exception as ex:
                print(str(ex))
            if city is not None and isinstance(city,City):
                if showProgress:
                    progress.next()
                if logFile:
                    print(f"{i:4d}/{count:4d}{rsum:6d}/{total:5d}({percent:4.1f}%)✅:{locationText}({locationCount})→{city} ({city.pop})",file=logFile)
                events=eventsByLocation[locationText]
                # loop over all events
                for event in events:
                    event.city=city.name
                    event.cityWikidataid=city.wikidataid
                    
                    event.region=city.region.name
                    event.regionIso=city.region.iso
                    event.regionWikidataid=city.region.wikidataid
                    event.country=city.country.name
                    event.countryIso=city.country.iso
                    event.countryWikidataid=city.country.wikidataid
            else:
                if logFile:
                    print(f"{i:4d}/{count:4d}{rsum:6d}/{total:5d}({percent:4.1f}%)❌:{locationText}({locationCount})",file=logFile)
                problems.append(locationText)
        problemMsg=f"found {len(problems)} problems"
        progress.done()
        for i,problem in enumerate(problems):
            if logFile:
                print(f"{i:4d}:{problem}",file=logFile)        
        print(problemMsg,file=logFile)      
        if addLocationInfo:
            eventManager.store()   
        


'''
Created on 2022-04-03

@author: wf
'''



class EventSeriesCompletion(object):
    '''
    complete a series of events from different datasources 
    '''
    tokenizer=None
    
    def __init__(self):
        '''
        Constructor
        '''            
    
    
        


'''
Created on 2021-05-31

@author: wf
'''
from collections import Counter
from lodstorage.tabulateCounter import TabulateCounter

class Tokenizer(object):
    '''
    categorize some text input
    ''' 
    def __init__(self,categories):
        self.categories=categories
        for category in categories:
            for neededFunc in ["checkMatch","itemFunc"]:
                hasNeededFunc=hasattr(category,neededFunc) and callable(getattr(category,neededFunc))
                if not hasNeededFunc:
                    raise Exception(f"category {category.name} has no {neededFunc} function")
        pass
    
    def tokenize(self,text,item):
        '''
        tokenize the given text for the given item
        '''
        tokenSequence=TokenSequence(text)
        tokenSequence.match(self.categories,item)
        return tokenSequence
        
class TokenSequence(object):
    '''
    a sequence of tokens
    '''
    
    def __init__(self,text,separator=' '):
        if text:
            self.words=text.split(separator)
        else:
            self.words=[]
        self.pos=-1
        self.matchResults=[]
        
    def next(self):
        while self.pos+1<len(self.words):
            self.pos+=1
            yield self.words[self.pos]
            
    def match(self,categories:list,item:object)->list:
        '''
        match me for the given categories
        
        '''
        self.item=item
        for tokenStr in self.next():
            for category in categories:
                if category.checkMatch(tokenStr):
                    token=Token(category,self,self.pos,tokenStr,item)
                    self.matchResults.append(token)
        return self.matchResults
    
    def __str__(self):
        text=f"{self.name}:{self.matchResults}"
        return text
        
        
class Token(object):
    '''
    a single categorized token
    '''
    
    def __init__(self,category,tokenSequence,pos,tokenStr,item):
        self.category=category
        self.name=category.name
        self.tokenSequence=tokenSequence
        self.pos=pos
        self.tokenStr=tokenStr
        self.value=category.add(item,tokenStr)
        
    def __str__(self):
        text=self.tokenStr
        return text
       
        
class Category(object):
    '''
    I am a category
    '''

    def __init__(self, name, itemFunc):
        '''
        Constructor
        '''
        self.name=name
        self.items={}
        self.itemFunc=itemFunc
        self.counter=Counter()
        self.subCategories={}
        
    def addCategory(self,category):
        self.subCategory[category.name]=category
        
    def add(self,item,propValue):
        '''
        add the given item with the given value
        '''
        value=self.itemFunc(propValue)
        if value in self.items:
            self.items[value].append(item)
        else:
            self.items[value]=[item]
        self.counter[value]+=1
        return value
        
   
    def mostCommonTable(self,tablefmt='pretty',limit=50):
        '''
        get the most common Table
        '''
        tabulateCounter=TabulateCounter(self.counter)
        table=tabulateCounter.mostCommonTable(tablefmt=tablefmt, limit=limit)
        return table
        